divert(_NullStream)
dnl -----------------------------------------------
dnl   m4 file for linux specific taxi features
dnl -----------------------------------------------

dnl note convention: initial lowercase after _ indicates function
dnl                  initial uppercase after _ indicates variable

dnl ###########################
dnl   list of global variables
dnl ###########################

define(`_Os', `linux')dnl
dnl _FopsPrototypes
dnl _FopsDefinitions
define(`_Success', `0')

dnl ###########################
dnl   error codes
dnl ###########################
define(`_ENOMEM', `-ENOMEM')

dnl ###########################
dnl   flags
dnl ###########################
define(`_WAITOK', `GFP_KERNEL')dnl tells malloc it is ok to wait for memory
define(`_NOWAIT', `GFP_ATOMIC')dnl tells malloc it is NOT ok to wait for memory

dnl ###########################
dnl   internal functions
dnl ###########################

dnl _addFopsPrototype(prototype)
dnl	adds the specified prototype to the list of prototypes in _FopsPrototypes
define(`_addFopsPrototype',dnl
`define(`_FopsPrototypes',dnl
defn(`_FopsPrototypes')
$1)')

dnl _addFile_ops_struct(method, methodName)
dnl     adds specified method and method name to file_operations data structure
define(`_addFile_ops_struct',dnl
`define(`_file_operations_list',dnl
defn(`_file_operations_list')dnl
`  $1:     $2,'
)')dnl

dnl _declareFopsPrototypes is a high level macro which emits the prototypes for all
dnl	file operation functions which have been registered
define(`_declareFopsPrototypes', 
`/* Prototypes for registered file operation functions */
defn(`_FopsPrototypes')');

dnl _addFopsDefinition(definition)
dnl	adds the specified definition to the list of definitions in _FopsDefinitions
define(`_addFopsDefinition',dnl
`define(`_FopsDefinitions',dnl
defn(`_FopsDefinitions')
$1)')

dnl _ModuleEntryFunctionStream       Represents the stream which will buffer access point initialization
dnl                                  statements at the end of module source code on Linux
dnl                                  e.g. void __init myModule_init(void) {...}
dnl                                       ...other code here...
dnl                                       module_init(myModule_init); //output this typically at end
dnl                                       EOF
define(`_ModuleEntryFunctionStream',`1')

dnl ########################################################################################
dnl _OsDefines defines all CPP symbols and vars that the target OS requires
dnl ########################################################################################
define(`_OsDefines',`
_ifndefDefine(`__KERNEL__')
_ifndefDefine(`__MODULE__')
');

dnl _addOsDefine(define)
define(`_addOsDefine',dnl
`define(`_OsDefines',dnl
defn(`_OsDefines')
$1)')

dnl ########################################################################################
dnl _osIncludes causes all pertinent OS specific headers to be included
dnl 	CHANGE: at some point this macro needs to do something EXTREEMLY clever:
dnl 	it needs to read the input file and find out which files must be included
dnl	perhaps by detecting which symbols are used.  
dnl	this will probably be subtle, and quick to anger.
dnl ########################################################################################
define(`_osIncludes',`dnl
`#'include <linux/kernel.h>
`#'include <linux/module.h>
`#'include <linux/config.h>
`#'include <linux/types.h>
`#'include <linux/fs.h>
`#'include <linux/slab.h>
`#'include <linux/mm.h>
`#'include <asm/uaccess.h>');

dnl ########################################################################################
dnl _osVars causes all pertinent OS specific variables to be declared
dnl ########################################################################################
define(`_osVars',dnl
`/* no OS specific vars for linux yet */
');

dnl ##########################################################################################
dnl _declareFopsStruct declares a file operations struct to be used in the driver's .c source
dnl 	the format of such a struct is obviously OS dependant.  under linux it comes from fs.h
dnl ##########################################################################################
define(`_declareFopsStruct',`
/*************************************************************************
 * File operations structure declaration.
 * Function pointers are automatically set to names of functions as the
 * functions are registered in the m4 source.
 *************************************************************************/
struct file_operations _DriverName`'_fops = {
_file_operations_list};
')

dnl _defineFops is a high level macro which spits out the definitions for all
dnl	file operation functions which have been registered
define(`_defineFops',
`/* Definitions for registered file operation functions */
defn(`_FopsDefinitions')');

dnl *********************************************************************
dnl *********************************************************************
dnl		    EXTERNALLY AVAILABLE FUNCTIONS
dnl *********************************************************************
dnl *********************************************************************

dnl ############################################################
dnl   init/exit function declarations
dnl   NOTE: intentional missing dnl's between divert statements
dnl ############################################################

dnl ##################################################
dnl   _load
dnl        > registers the function used to load
dnl          the module when requested by OS
dnl ##################################################
define(`_load',`dnl
define(`_LoadFN', `_DriverName`'_init')dnl
define(`_LoadProto',`static int __init _LoadFN`()'')dnl
divert(_ModuleEntryFunctionStream)dnl
module_init(_LoadFN);
divert(_StdoutStream)dnl
define(`_defineInitModule',`dnl
_LoadProto {
  int _MethodRetval = _Success;
  _DeclareTaxiRetval
$1}
')dnl
')dnl

dnl ##################################################
dnl   _unload
dnl        > registers the function used to unload
dnl          the module when requested by OS
dnl ##################################################
define(`_unload',`dnl
define(`_UnloadFN', `_DriverName`'_exit')dnl
define(`_UnloadProto',`static void __exit _UnloadFN`()'')dnl
divert(_ModuleEntryFunctionStream)dnl
module_exit(_UnloadFN);
divert(_StdoutStream)dnl
define(`_defineExitModule',`dnl
_UnloadProto {
$1}
')dnl
')dnl

dnl #################################################
dnl FOR NOW, NOT EXPLICITLY NEEDED BECAUSE ON EOF, ALL STREAMS
dnl ARE OUTPUT TO STDOUT.  HOWEVER, LATER ON, IF WE USE 
dnl MULTIPLE DIVERSIONS, IT MAY BE USEFUL TO CONTROL THE
dnl ORDER IN WHICH THEY ARE OUTPUTTED
dnl
dnl undivert(_ModuleEntryFunctionStream)dnl
dnl #################################################

dnl _lseek(`codeBlock')  -registers a file operation function
define(`_lseek',dnl
`define(`_FopLSEEKFN', `_DriverName`'_LSEEK')dnl
define(`_FopLSEEKProto',dnl
`static loff_t _FopLSEEKFN`'(struct file *filp, loff_t offset, int orig)')dnl
define(`_FilePosition', `filp->f_pos')dnl
_addFopsPrototype(_FopLSEEKProto;)dnl
_addFopsDefinition(_FopLSEEKProto {$1})dnl
_addFile_ops_struct(`llseek', `_DriverName`'_LSEEK')dnl
')

dnl _read(`codeBlock')  -registers a file operation function
define(`_read',dnl
`define(`_FopREADFN', `_DriverName`'_READ')dnl
define(`_FopREADProto',dnl
`static ssize_t _FopREADFN`'(struct file *filp, char *buf, size_t count, loff_t *f_pos)')dnl # note, no semi-colon!
define(`_InputBase', `buf')dnl
define(`_InputLength', `count')dnl
define(`_FilePosition', `filp->f_pos')dnl
_addFopsPrototype(_FopREADProto;)dnl
_addFopsDefinition(_FopREADProto {$1})dnl
_addFile_ops_struct(`read', `_DriverName`'_READ')dnl
')

dnl _setWriteBytesWritten is a helper function for _write
define(`_setWriteBytesWritten', `define(`_WriteBytesWritten', `$1')')
dnl _write(`codeBlock')  -registers a file operation function
define(`_write',dnl
`define(`_FopWRITEFN', `_DriverName`'_WRITE')dnl
define(`_FopWRITEProto',dnl
`static ssize_t _FopWRITEFN`'(struct file *filp, const char *buf, size_t count, loff_t *f_pos)')dnl
define(`_InputBase', `buf')dnl
define(`_InputLength', `count')dnl
_addFopsPrototype(_FopWRITEProto;)dnl
_addFopsDefinition(`_FopWRITEProto {
  _DeclareTaxiRetval
  int _MethodRetval = _Success;
$1
  _MethodRetval = _WriteBytesWritten;
  _return(_MethodRetval)
`}'')dnl
_addFile_ops_struct(`write', `_DriverName`'_WRITE')dnl
')

dnl _open(`codeBlock')  -registers a file operation function
define(`_open',dnl
`define(`_FopOPENFN', `_DriverName`'_OPEN')dnl
define(`_FopOPENProto',dnl
`static ssize_t _FopOPENFN`'(struct inode *node, struct file *filp)')dnl # note, no semi-colon!
_addFopsPrototype(_FopOPENProto;)dnl
_addFopsDefinition(_FopOPENProto {$1})dnl
_addFile_ops_struct(`open', `_DriverName`'_OPEN')dnl
')

dnl _release(`codeBlock')  -registers a file operation function
define(`_release',dnl
`define(`_FopRELEASEFN', `_DriverName`'_RELEASE')dnl
define(`_FopRELEASEProto',dnl
`static ssize_t _FopRELEASEFN`'(struct inode *node, struct file *filp)')dnl # note, no semi-colon!
_addFopsPrototype(_FopRELEASEProto;)dnl
_addFopsDefinition(_FopRELEASEProto {$1})dnl
_addFile_ops_struct(`release', `_DriverName`'_RELEASE')dnl
')

dnl _kernelPrint(`string')  -prints a string to the kernel logging facility
define(`_kernelPrint',`printk($1);')dnl

dnl _osLicense(`licenseType')  -emits OS specific license type macro
define(`_osLicense', `MODULE_LICENSE("_DriverLicenseType");')

dnl _setPortRange(`base',`range')  -allocates range of I/O starting at `base' for module
define(`_setPortRange',`//setPortRange definition')dnl

dnl _setMemoryRange(`base',`range')  -allocates range of I/O shared memory starting at `base' for module
define(`_setMemoryRange',`//setMemoryRange definition')dnl

dnl _unsetPortRange(`base',`range')  -allocates range of I/O starting at `base' for module
define(`_unsetPortRange', `//unsetPortRange definition')dnl

dnl _unsetMemoryRange(`base',`range')  -allocates range of I/O shared memory starting at `base' for module
define(`_unsetMemoryRange', `//unsetMemoryRange definition')dnl

dnl _markModuleUsage   -acquires semaphore preventing module removal
define(`_markModuleUsage', `MOD_INC_USE_COUNT;')

dnl _unmarkModuleUsage   -releases semaphore allowing module removal
define(`_unmarkModuleUsage', `MOD_DEC_USE_COUNT;')

dnl _registerModule()  -registers device with the OS
dnl	returns: the emitted c code returns the retval from the registration
define(`_registerModule',dnl
`_TaxiRetval = register_chrdev(_MajorNumber, "_DriverName", &`'_DriverName`'_fops);')dnl

dnl _unregisterModule()  -unregisters device with the OS
define(`_unregisterModule', `unregister_chrdev(_MajorNumber, "_DriverName");')dnl

dnl _copyin(userland address, kernelspace address, count)
dnl    copies a string of length count from userland to kernelspace
define(`_copyin', `_TaxiRetval = copy_from_user($2, $1, $3);')dnl

dnl _mallocDeclare(type)
dnl	declares a malloc type
define(`_mallocDeclare', `dnl')dnl does nothing under linux

dnl _mallocDefine(type, shortdesc, longdesc)
dnl	defines a malloc type
define(`_mallocDefine', `dnl')dnl

dnl _malloc(size, type, flags)
dnl	allocates memory of the appropriate size, and cast
define(`_malloc', `kmalloc($1, $3)')dnl $2 not used under linux

dnl _free(addr, type)
dnl	frees memory allocated with _malloc
define(`_free', `kfree($1)')dnl $2 not used under linux

dnl _requestIrq(irq, handler, flags)
dnl	request that handler be added to the interrupt vector for given irq.
define(`_requestIrq', `request_irq($1, $2, $3, "_DriverName", NULL)')

divert(_StdoutStream)
